"""DICOM dictionires related utilities."""
import os
import re
import typing as t

from fw_utils import AnyFile, open_any
from pydicom.datadict import add_private_dict_entry
from pydicom.datadict import private_dictionaries as pydicom_private_dict

__all__ = ["load_dcmdict"]

# See: https://github.com/DCMTK/dcmtk/blob/DCMTK-3.6.6/dcmdata/data/private.dic
DCMTK_LINE_RE = (
    r"^\("
    r"(?P<group>[0-9a-f]{4})(-o-(?P<group_end>[0-9a-f]{4}))?,"
    r'"(?P<creator>[^"]+)",'
    r"(?P<elem>[0-9a-f]{2,4})"
    r"\)"
    r"\t(?P<vr>[A-Z]{2})"
    r"\t(?P<name>[^\t]+)"
    r"\t(?P<vm>[^\t]+)"
)


def load_dcmdict(file: AnyFile) -> None:
    """Parse and load a DCMTK-style private dict file, extending pydicom's dict."""
    with open_any(file) as file_obj:
        for line in file_obj:
            match = re.match(DCMTK_LINE_RE, line)
            if not match:
                continue

            group = int(match.group("group"), 16)
            group_end = int(match.group("group_end") or "0", 16) or group
            creator = match.group("creator")
            elem = int(match.group("elem"), 16)
            vr = match.group("vr")
            name = match.group("name")
            vm = match.group("vm")

            for grp in range(group, group_end + 1):
                if grp % 2 == 0:
                    continue
                tag = grp << 16 | elem
                add_private_dict_entry(creator, tag, vr, name, vm)


def load_private_dictionaries() -> None:
    """Load shipped private dictionary and DCMDICTPATH if set in the env."""
    # load the shipped private dict below 1st
    for creator, entries in private_dict.items():
        pydicom_private_dict.setdefault(creator, {}).update(entries)

    # DCMDICTPATH loaded 2nd to enable more dynamic overrides
    dcm_dict_path = os.getenv("DCMDICTPATH")
    if dcm_dict_path:
        load_dcmdict(dcm_dict_path)  # pragma: no cover


private_dict: t.Dict[str, t.Dict[str, t.Tuple[str, str, str, str]]] = {
    "AGFA PACS Archive Mirroring 1.0": {
        "0031xx00": ("CS", "1", "StudyStatus", ""),
        "0031xx01": ("UL", "1", "DateTimeVerified", ""),
    },
    "CARDIO-D.R. 1.0": {"0019xx30": ("UL", "1", "MaximumImageFrameSize", "")},
    "CMR42 CIRCLECVI": {
        "0025xx10": ("LO", "1", "WorkspaceID", ""),
        "0025xx20": ("LO", "1", "WorkspaceTimeString", ""),
        "0025xx30": ("OB", "1", "WorkspaceStream", ""),
    },
    "DCMTK_ANONYMIZER": {
        "0009xx00": ("SQ", "1", "AnonymizerUIDMap", ""),
        "0009xx10": ("UI", "1", "AnonymizerUIDKey", ""),
        "0009xx20": ("UI", "1", "AnonymizerUIDValue", ""),
        "0009xx30": ("SQ", "1", "AnonymizerPatientIDMap", ""),
        "0009xx40": ("LO", "1", "AnonymizerPatientIDKey", ""),
        "0009xx50": ("LO", "1", "AnonymizerPatientIDValue", ""),
    },
    "DLX_EXAMS_01": {
        "0015xx01": ("DS", "1", "StenosisCalibrationRatio", ""),
        "0015xx02": ("DS", "1", "StenosisMagnification", ""),
        "0015xx03": ("DS", "1", "CardiacCalibrationRatio", ""),
    },
    "FDMS 1.0": {
        "0019xx60": ("SH", "1", "RadiographersCode", ""),
        "0029xx34": ("US", "1", "MagnificationReductionRatio", ""),
    },
    "Flywheel": {
        "0021xx00": ("LO", "1", "DICOM Send", ""),
    },
    "GEMS_ACQU_01": {
        "0019xx02": ("SL", "1", "NumberOfCellsInDetector", ""),
        "0019xx0f": ("DS", "1", "HorizontalFrameOfReference", ""),
        "0019xx1b": ("DS", "1", "LastScanLocation", ""),
        "0019xx23": ("DS", "1", "TableSpeed", ""),
        "0019xx24": ("DS", "1", "MidScanTime", ""),
        "0019xx26": ("SL", "1", "DegreesOfAzimuth", ""),
        "0019xx27": ("DS", "1", "GantryPeriod", ""),
        "0019xx39": ("SS", "1", "ScanFOVType", ""),
        "0019xx43": ("SS", "1", "TotalSegmentsRequested", ""),
        "0019xx4a": ("SS", "1", "TotalNumberOfRefChannels", ""),
        "0019xx52": ("SS", "1", "ReconPostProcessingFlag", ""),
        "0019xx5e": ("SL", "1", "NumberOfChannels1To512", ""),
        "0019xx6a": ("SS", "1", "DependantOnNumberOfViewsProcessed", ""),
        "0019xx72": ("DS", "1", "ZChannelAvgOverViews", ""),
        "0019xx73": ("DS", "1", "AvgOfLeftRefChannelsOverViews", ""),
        "0019xx74": ("DS", "1", "MaxLeftChannelOverViews", ""),
        "0019xx75": ("DS", "1", "AvgOfRightRefChannelsOverViews", ""),
        "0019xx76": ("DS", "1", "MaxRightChannelOverViews", ""),
        "0019xx7e": ("SS", "1", "NumberOfEchos", ""),
        "0019xx8f": ("SS", "1", "SwapPhaseFrequency", ""),
        "0019xx91": ("DS", "1", "PulseTime", ""),
        "0019xx92": ("SL", "1", "SliceOffsetOnFrequencyAxis", ""),
        "0019xx93": ("DS", "1", "CenterFrequency", ""),
        "0019xx94": ("SS", "1", "TransmitGain", ""),
        "0019xx95": ("SS", "1", "AnalogReceiverGain", ""),
        "0019xx96": ("SS", "1", "DigitalReceiverGain", ""),
        "0019xx98": ("SS", "1", "CenterFrequencyMethod", ""),
        "0019xx9f": ("SS", "1", "TransmittingCoil", ""),
        "0019xxc1": ("SS", "1", "SurfaceCoilIntensityCorrectionFlag", ""),
        "0019xxcb": ("SS", "1", "PrescribedFlowAxis", ""),
        "0019xxd3": ("SH", "1", "ProjectionAlgorithm", ""),
        "0019xxd7": ("SS", "1", "CardiacPhases", ""),
        "0019xxd9": ("DS", "1", "ConcatenatedSAT", ""),
        "0019xxdf": ("DS", "1", "UserData23", ""),
        "0019xxe0": ("DS", "1", "UserData24", ""),
    },
    "GEMS_PARM_01": {
        "0043xx08": ("SS", "1", "RespiratoryRateInBPM", ""),
        "0043xx0b": ("DS", "1", "PeakRateOfChangeOfGradientField", ""),
        "0043xx0c": ("DS", "1", "LimitsInUnitsOfPercent", ""),
        "0043xx13": ("SS", "5", "ReconKernelParameters", ""),
        "0043xx1e": ("DS", "1", "DeltaStartTime", ""),
        "0043xx26": ("US", "6", "NoViewsRefChannelsBlocked", ""),
        "0043xx28": ("OB", "1", "UniqueImageIdentifier", ""),
        "0043xx2d": ("SH", "1", "StringSlopField1", ""),
        "0043xx2f": ("SS", "1", "RawDataType", ""),
        "0043xx30": ("SS", "1", "RawDataType", ""),
        "0043xx31": ("DS", "2", "RACoordOfTargetReconCentre", ""),
        "0043xx32": ("SS", "1", "RawDataType", ""),
        "0043xx38": ("FL", "24", "User25ToUser48", ""),
        "0043xx39": ("IS", "4", "SlopInteger6ToSlopInteger9", ""),
        "0043xx60": ("IS", "8", "SlopInteger10ToSlopInteger17", ""),
        "0043xx6f": ("DS", "3", "ScannerTableEntry", ""),
        "0043xx74": ("US", "1", "NumberOfRestVolumes", ""),
        "0043xx75": ("US", "1", "NumberOfActiveVolumes", ""),
        "0043xx76": ("US", "1", "NumberOfDummyScans", ""),
        "0043xx7f": ("DS", "1-n", "EDWIScaleFactor", ""),
        "0043xx87": ("UT", "1", "ScannerSoftwareVersionLongForm", ""),
    },
    "GEMS_RELA_01": {
        "0021xx56": ("SL", "1", "IntegerSlop", ""),
        "0021xx57": ("SL", "1", "IntegerSlop", ""),
        "0021xx58": ("SL", "1", "IntegerSlop", ""),
        "0021xx59": ("SL", "1", "IntegerSlop", ""),
        "0021xx5a": ("SL", "1", "IntegerSlop", ""),
        "0021xx5b": ("DS", "1", "FloatSlop", ""),
        "0021xx5c": ("DS", "1", "FloatSlop", ""),
        "0021xx5d": ("DS", "1", "FloatSlop", ""),
        "0021xx5e": ("DS", "1", "FloatSlop", ""),
        "0021xx5f": ("DS", "1", "FloatSlop", ""),
    },
    "GEMS_SENO_02": {
        "0045xx06": ("DS", "1", "Angulation", ""),
        "0045xx0d": ("DS", "1", "ROIOriginXY", ""),
        "0045xx11": ("DS", "2", "ReceptorSizeCmXY", ""),
        "0045xx12": ("IS", "2", "ReceptorSizePixelsXY", ""),
        "0045xx16": ("IS", "2", "BinningFactorXY", ""),
        "0045xx20": ("DS", "1", "MeanOfRegionGrayLevels", ""),
        "0045xx28": ("CS", "1", "WindowingType", ""),
        "0045xx2a": ("IS", "1", "CrosshairCursorXCoordinates", ""),
        "0045xx2b": ("IS", "1", "CrosshairCursorYCoordinates", ""),
    },
    "GEMS_SERS_01": {"0025xx14": ("SL", "1", "IndicatesNumberOfUpdatesToHeader", "")},
    "GEMS_STDY_01": {
        "0023xx74": ("SL", "1", "NumberOfUpdatesToHeader", ""),
        "0023xx7d": ("SS", "1", "IndicatesIfStudyHasCompleteInfo", ""),
    },
    "GE_GENESIS_REV3.0": {
        "0019xx8f": ("SS", "1", "SwapPhaseFrequency", ""),
        "0019xxa4": ("SS", "1", "SATFatWaterBone", ""),
        "0019xxc1": ("SS", "1", "SurfaceCoilIntensityCorrectionFlag", ""),
        "0043xx27": ("SH", "1", "ScanPitchRatio", ""),
    },
    "INTELERAD MEDICAL SYSTEMS": {
        "0029xx01": ("FD", "1", "ImageCompressionFraction", ""),
        "0029xx02": ("FD", "1", "ImageQuality", ""),
        "0029xx03": ("FD", "1", "ImageBytesTransferred", ""),
        "0029xx10": ("SH", "1", "J2cParameterType", ""),
        "0029xx11": ("US", "1", "J2cPixelRepresentation", ""),
        "0029xx12": ("US", "1", "J2cBitsAllocated", ""),
        "0029xx13": ("US", "1", "J2cPixelShiftValue", ""),
        "0029xx14": ("US", "1", "J2cPlanarConfiguration", ""),
        "0029xx15": ("DS", "1", "J2cRescaleIntercept", ""),
        "0029xx20": ("LO", "1", "PixelDataMD5SumPerFrame", ""),
        "0029xx21": ("US", "1", "HistogramPercentileLabels", ""),
        "0029xx22": ("FD", "1", "HistogramPercentileValues", ""),
        "3f01xx06": ("LO", "1", "OrderGroupNumber", ""),
        "3f01xx07": ("SH", "1", "StrippedPixelData", ""),
        "3f01xx08": ("SH", "1", "PendingMoveRequest", ""),
    },
    "OCULUS Optikgeraete GmbH": {
        "0029xx10": ("OB", "1", "OriginalMeasuringData", ""),
        "0029xx12": ("UL", "1", "OriginalMeasuringDataLength", ""),
        "0029xx20": ("OB", "1", "OriginalMeasuringRawData", ""),
        "0029xx22": ("UL", "1", "OriginalMeasuringRawDataLength", ""),
    },
    "Philips EV Imaging DD022": {"2007xx12": ("LO", "1", "VolumeSequenceCapture", "")},
    "Philips Imaging DD 001": {
        "2001xx05": ("SS", "1", "GraphicAnnotationParentID", ""),
        "2001xx09": ("FL", "1", "ImagePrepulseDelay", ""),
        "2001xx0c": ("CS", "1", "ArrhythmiaRejection", ""),
        "2001xx0e": ("CS", "1", "CardiacCycled", ""),
        "2001xx0f": ("SS", "1", "CardiacGateWidth", ""),
        "2001xx1e": ("CS", "1", "MRSeriesReformatAccuracy", ""),
        "2001xx3d": ("UL", "1", "ContourFillColor", ""),
        "2001xx3f": ("CS", "1", "ZoomMode", ""),
        "2001xx43": ("IS", "1", "EllipsDisplShutMajorAxFrstEndPnt", ""),
        "2001xx44": ("IS", "1", "EllipsDisplShutMajorAxScndEndPnt", ""),
        "2001xx45": ("IS", "1", "EllipsDisplShutOtherAxFrstEndPnt", ""),
        "2001xx50": ("LO", "1", "GraphicMarkerType", ""),
        "2001xx51": ("IS", "1", "OverlayPlaneID", ""),
        "2001xx52": ("UI", "1", "ImagePresentationStateUID", ""),
        "2001xx54": ("FL", "1", "ContourFillTransparency", ""),
        "2001xx58": ("UL", "1", "ContrastTransferTaste", ""),
        "2001xx61": ("CS", "1", "SeriesTransmitted", ""),
        "2001xx62": ("CS", "1", "SeriesCommitted", ""),
        "2001xx65": ("SQ", "1", "ROIOverlayPlane", ""),
        "2001xx67": ("CS", "1", "LinearPresentationGLTrafoShapeSub", ""),
        "2001xx68": ("SQ", "1", "LinearModalityGLTrafo", ""),
        "2001xx6a": ("SQ", "1", "SpatialTransformation", ""),
        "2001xx71": ("CS", "1", "GraphicConstraint", ""),
        "2001xx72": ("IS", "1", "EllipsDisplShutOtherAxScndEndPnt", ""),
        "2001xx76": ("UI", "1", "MRNumberOfFrames", ""),
        "2001xx7a": ("FL", "1", "WindowRoundingFactor", ""),
        "2001xx7c": ("US", "1", "FrameNumber", ""),
        "2001xx86": ("IS", "1", "MRSeriesNrOfPhaseEncodingSteps", ""),
        "2001xx8b": ("SH", "1", "MRSeriesTransmittingCoil", ""),
        "2001xx90": ("LO", "1", "TextForegroundColor", ""),
        "2001xx91": ("LO", "1", "TextBackgroundColor", ""),
        "2001xx92": ("LO", "1", "TextShadowColor", ""),
        "2001xx9c": ("LO", "1", "GraphicAnnotationLabel", ""),
    },
    "Philips MR Imaging DD 001": {
        "2005xx00": ("FL", "1", "MRImageAngulationAP", ""),
        "2005xx01": ("FL", "1", "MRImageAngulationFH", ""),
        "2005xx02": ("FL", "1", "MRImageAngulationRL", ""),
        "2005xx03": ("IS", "1", "ImageAnnotationCount", ""),
        "2005xx04": ("CS", "1", "MRImageDisplayOrientation", ""),
        "2005xx07": ("IS", "1", "ImageLineCount", ""),
        "2005xx08": ("FL", "1", "MRImageOffCentreAP", ""),
        "2005xx09": ("FL", "1", "MRImageOffCentreFH", ""),
        "2005xx0a": ("FL", "1", "MRImageOffCentreRL", ""),
        "2005xx0b": ("FL", "1", "MRMaxFP", ""),
        "2005xx0c": ("FL", "1", "MRMinFP", ""),
        "2005xx0d": ("FL", "1", "MRScaleIntercept", ""),
        "2005xx0e": ("FL", "1", "MRScaleSlope", ""),
        "2005xx11": ("CS", "1", "MRImageTypeMR", ""),
        "2005xx12": ("CS", "1", "MRCardiacGating", ""),
        "2005xx13": ("CS", "1", "MRSeriesDevelopmentMode", ""),
        "2005xx14": ("CS", "1", "MRSeriesDiffusion", ""),
        "2005xx15": ("CS", "1", "MRFatSaturation", ""),
        "2005xx16": ("CS", "1", "MRFlowCompensation", ""),
        "2005xx17": ("CS", "1", "MRFourierInterpolation", ""),
        "2005xx18": ("LO", "1", "MRHardcopyProtocol", ""),
        "2005xx19": ("CS", "1", "MRInverseReconstructed", ""),
        "2005xx1a": ("SS", "1", "MRLabelSyntax", ""),
        "2005xx1b": ("CS", "1", "MRMagnetiPrepared", ""),
        "2005xx1c": ("CS", "1", "MRMagnetTransferConst", ""),
        "2005xx1d": ("SS", "1", "MRMeasurementScanResolution", ""),
        "2005xx1e": ("SH", "1", "MIPProtocol", ""),
        "2005xx1f": ("SH", "1", "MPRProtocol", ""),
        "2005xx21": ("SS", "1", "MRNumberOfMixes", ""),
        "2005xx22": ("IS", "1", "MRNumberOfReferences", ""),
        "2005xx23": ("SS", "1", "MRNumberOfSlabs", ""),
        "2005xx25": ("SS", "1", "MRNumberOfVolumes", ""),
        "2005xx26": ("CS", "1", "MROverSampleingPhase", ""),
        "2005xx27": ("CS", "1", "MRPackageMode", ""),
        "2005xx28": ("CS", "1", "MRPartialFourierFrequency", ""),
        "2005xx29": ("CS", "1", "MRPartialFourierPhase", ""),
        "2005xx2a": ("IS", "1", "MRPatientReferenceID", ""),
        "2005xx2b": ("SS", "1", "MRPercentScanComplete", ""),
        "2005xx2c": ("CS", "1", "MRPhaseEncodedRecording", ""),
        "2005xx2d": ("SS", "1", "NumberOfStackSlices", ""),
        "2005xx2e": ("CS", "1", "MRPPGPPUGating", ""),
        "2005xx2f": ("CS", "1", "MRSpatialPresaturation", ""),
        "2005xx31": ("CS", "1", "MRRespiratoryGating", ""),
        "2005xx32": ("CS", "1", "SampleRepresentation", ""),
        "2005xx34": ("CS", "1", "MRSegmentedKSpace", ""),
        "2005xx35": ("CS", "1", "MRSeriesDataType", ""),
        "2005xx36": ("CS", "1", "MRSeriesIsCardiac", ""),
        "2005xx37": ("CS", "1", "MRSeriesIsSpectro", ""),
        "2005xx38": ("CS", "1", "MRSpoiled", ""),
        "2005xx39": ("CS", "1", "MRSteadyState", ""),
        "2005xx3a": ("SH", "1", "MRSubAnatomy", ""),
        "2005xx3b": ("CS", "1", "MRTimeReversedSteadyState", ""),
        "2005xx3c": ("CS", "1", "MRSeriesTone", ""),
        "2005xx3d": ("SS", "1", "MRNumberOfRRIntervalRanges", ""),
        "2005xx3e": ("SL", "1", "MRRRIntervalsDistribution", ""),
        "2005xx3f": ("SL", "1", "MRPlanScanAcquisitionNo", ""),
        "2005xx40": ("SL", "1", "MRChemicalShiftNo", ""),
        "2005xx41": ("SL", "1", "MRPlanScanDynamicScanNo", ""),
        "2005xx42": ("SL", "1", "MRPlanScanSurveyEchoNo", ""),
        "2005xx43": ("CS", "1", "MRPlanScanImageType", ""),
        "2005xx44": ("SL", "1", "MRPlanScanPhaseNo", ""),
        "2005xx45": ("SL", "1", "MRPlanScanReconstructionNo", ""),
        "2005xx46": ("CS", "1", "MRPlanScanScanningSequence", ""),
        "2005xx47": ("SL", "1", "MRPlanScanSliceNo", ""),
        "2005xx48": ("IS", "1", "MRReferenceAcquisitionNo", ""),
        "2005xx49": ("IS", "1", "MRReferenceChemicalShiftNo", ""),
        "2005xx4a": ("IS", "1", "MRReferenceDynamicScanNo", ""),
        "2005xx4b": ("IS", "1", "MRReferenceEchoNo", ""),
        "2005xx4c": ("CS", "1", "MRReferenceEntity", ""),
        "2005xx4d": ("CS", "1", "MRReferenceImageType", ""),
        "2005xx4e": ("FL", "1", "MRSlabFovRL", ""),
        "2005xx4f": ("FL", "1", "MRSlabOffcentreAP", ""),
        "2005xx50": ("FL", "1", "MRSlabOffcentreFH", ""),
        "2005xx51": ("FL", "1", "MRSlabOffcentreRL", ""),
        "2005xx52": ("CS", "1", "MRSlabType", ""),
        "2005xx53": ("CS", "1", "MRSlabViewAxis", ""),
        "2005xx54": ("FL", "1", "MRVolumeAngulationAP", ""),
        "2005xx55": ("FL", "1", "MRVolumeAngulationFH", ""),
        "2005xx56": ("FL", "1", "MRVolumeAngulationRL", ""),
        "2005xx57": ("FL", "1", "MRVolumeFovAP", ""),
        "2005xx58": ("FL", "1", "MRVolumeFovFH", ""),
        "2005xx59": ("FL", "1", "MRVolumeFovRL", ""),
        "2005xx5a": ("FL", "1", "MRVolumeOffcentreAP", ""),
        "2005xx5b": ("FL", "1", "MRVolumeOffcentreFH", ""),
        "2005xx5c": ("FL", "1", "MRVolumeOffcentreRL", ""),
        "2005xx5d": ("CS", "1", "MRVolumeType", ""),
        "2005xx5e": ("CS", "1", "MRVolumeViewAxis", ""),
        "2005xx5f": ("CS", "1", "MRStudyOrigin", ""),
        "2005xx60": ("IS", "1", "MRStudySequenceNumber", ""),
        "2005xx61": ("CS", "1", "MRImagePrepulseType", ""),
        "2005xx63": ("SS", "1", "MRfMRIStatusIndication", ""),
        "2005xx64": ("IS", "1", "MRReferencePhaseNo", ""),
        "2005xx65": ("IS", "1", "MRReferenceReconstructionNo", ""),
        "2005xx66": ("CS", "1", "MRReferenceScanningSequence", ""),
        "2005xx67": ("IS", "1", "MRReferenceSliceNo", ""),
        "2005xx68": ("CS", "1", "MRReferenceType", ""),
        "2005xx69": ("FL", "1", "MRSlabAngulationAP", ""),
        "2005xx6a": ("FL", "1", "MRSlabAngulationFH", ""),
        "2005xx6b": ("FL", "1", "MRSlabAngulationRL", ""),
        "2005xx6c": ("FL", "1", "MRSlabFovAP", ""),
        "2005xx6d": ("FL", "1", "MRSlabFovFH", ""),
        "2005xx6e": ("CS", "1", "MRImageScanningSequencePrivate", ""),
        "2005xx6f": ("CS", "1", "MRSeriesAcquisitionTypePrivate", ""),
        "2005xx70": ("LO", "1", "MRSeriesHardcopyProtocolEV", ""),
        "2005xx71": ("FL", "1", "MRStackAngulationAP", ""),
        "2005xx72": ("FL", "1", "MRStackAngulationFH", ""),
        "2005xx73": ("FL", "1", "MRStackAngulationRL", ""),
        "2005xx74": ("FL", "1", "MRStackFovAP", ""),
        "2005xx75": ("FL", "1", "MRStackFovFH", ""),
        "2005xx76": ("FL", "1", "MRStackFovRL", ""),
        "2005xx78": ("FL", "1", "MRStackOffcentreAP", ""),
        "2005xx79": ("FL", "1", "MRStackOffcentreFH", ""),
        "2005xx7a": ("FL", "1", "MRStackOffcentreRL", ""),
        "2005xx7b": ("CS", "1", "MRStackPreparationDirection", ""),
        "2005xx7e": ("FL", "1", "MRStackSliceDistance", ""),
        "2005xx80": ("SQ", "1", "SeriesPlanScan", ""),
        "2005xx81": ("CS", "1", "MRStackViewAxis", ""),
        "2005xx84": ("SQ", "1", "SeriesReference", ""),
        "2005xx85": ("SQ", "1", "SeriesVolume", ""),
        "2005xx86": ("SS", "1", "MRNumberOfGeometry", ""),
        "2005xx87": ("SL", "1", "MRNumberOfGeometrySlices", ""),
        "2005xx88": ("FL", "1", "MRGeomAngulationAP", ""),
        "2005xx89": ("FL", "1", "MRGeomAngulationFH", ""),
        "2005xx8a": ("FL", "1", "MRGeomAngulationRL", ""),
        "2005xx8b": ("FL", "1", "MRGeomFOVAP", ""),
        "2005xx8c": ("FL", "1", "MRGeomFOVFH", ""),
        "2005xx8d": ("FL", "1", "MRGeomFOVRL", ""),
        "2005xx8e": ("FL", "1", "MRGeomOffCentreAP", ""),
        "2005xx8f": ("FL", "1", "MRGeomOffCentreFH", ""),
        "2005xx90": ("FL", "1", "MRGeomOffCentreRL", ""),
        "2005xx91": ("CS", "1", "MRGeomPreparationDirect", ""),
        "2005xx92": ("FL", "1", "MRGeomRadialAngle", ""),
        "2005xx93": ("CS", "1", "MRGeomRadialAxis", ""),
        "2005xx94": ("FL", "1", "MRGeomSliceDistance", ""),
        "2005xx95": ("SL", "1", "MRGeomSliceNumber", ""),
        "2005xx96": ("CS", "1", "MRGeomType", ""),
        "2005xx97": ("CS", "1", "MRGeomViewAxis", ""),
        "2005xx98": ("CS", "1", "MRGeomColour", ""),
        "2005xx99": ("CS", "1", "MRGeomApplicationType", ""),
        "2005xx9a": ("SL", "1", "MRGeomId", ""),
        "2005xx9b": ("SH", "1", "MRGeomApplicationName", ""),
        "2005xx9c": ("SH", "1", "MRGeomLableName", ""),
        "2005xx9d": ("CS", "1", "MRGeomLineStyle", ""),
        "2005xx9e": ("SQ", "1", "SeriesGeom", ""),
        "2005xx9f": ("CS", "1", "MRSeriesSpectralSelectiveExcitationPulse", ""),
    },
    "Philips MR Imaging DD 002": {
        "2005xx15": ("LO", "1", "MRUserName", ""),
        "2005xx16": ("LO", "1", "MRPassWord", ""),
        "2005xx17": ("LO", "1", "MRServerName", ""),
        "2005xx18": ("LO", "1", "MRDataBaseName", ""),
        "2005xx19": ("LO", "1", "MRRootName", ""),
        "2005xx20": ("LO", "1", "DMIApplicationName", ""),
        "2005xx2d": ("LO", "1", "MRRootId", ""),
        "2005xx32": ("SQ", "1", "MRBlobDataObjectArray", ""),
        "2005xx34": ("LT", "1", "SeriesTransactionUID", ""),
        "2005xx35": ("IS", "1", "ParentID", ""),
        "2005xx36": ("PN", "1", "ParentType", ""),
        "2005xx37": ("PN", "1", "MRBlobName", ""),
        "2005xx38": ("PN", "1", "MRApplicationName", ""),
        "2005xx39": ("PN", "1", "MRTypeName", ""),
        "2005xx40": ("PN", "1", "MRVersionStr", ""),
        "2005xx41": ("PN", "1", "MRCommentStr", ""),
        "2005xx42": ("CS", "1", "BlobInFile", ""),
        "2005xx43": ("SL", "1", "MRActualBlobSize", ""),
        "2005xx44": ("OW", "1", "MRBlobData", ""),
        "2005xx45": ("PN", "1", "BlobFilename", ""),
        "2005xx46": ("SL", "1", "BlobOffset", ""),
        "2005xx47": ("CS", "1", "MRBlobFlag", ""),
        "2005xx99": ("UL", "1", "MRNumberOfRequestExcerpts", ""),
    },
    "Philips MR Imaging DD 003": {
        "2005xx00": ("UL", "1", "MRNumberOfSOPCommon", ""),
        "2005xx01": ("UL", "1", "MRNoOfFilmConsumption", ""),
        "2005xx13": ("UL", "1", "MRNumberOfCodes", ""),
        "2005xx34": ("SL", "1", "MRNumberOfImagePerSeriesRef", ""),
        "2005xx43": ("SS", "1", "MRNoDateOfLastCalibration", ""),
        "2005xx44": ("SS", "1", "MRNoTimeOfLastCalibration", ""),
        "2005xx45": ("SS", "1", "MRNrOfSoftwareVersion", ""),
        "2005xx47": ("SS", "1", "MRNrOfPatientOtherNames", ""),
        "2005xx48": ("SS", "1", "MRNrOfReqRecipeOfResults", ""),
        "2005xx49": ("SS", "1", "MRNrOfSeriesOperatorsName", ""),
        "2005xx50": ("SS", "1", "MRNrOfSeriesPerfPhysiName", ""),
        "2005xx51": ("SS", "1", "MRNrOfStudyAdmittingDiagnosticDescr", ""),
        "2005xx52": ("SS", "1", "MRNrOfStudyPatientContrastAllergies", ""),
        "2005xx53": ("SS", "1", "MRNrOfStudyPatientMedicalAlerts", ""),
        "2005xx54": ("SS", "1", "MRNrOfStudyPhysiciansOfRecord", ""),
        "2005xx55": ("SS", "1", "MRNrOfStudyPhysiReadingStudy", ""),
        "2005xx56": ("SS", "1", "MRNrSCSoftwareVersions", ""),
        "2005xx57": ("SS", "1", "MRNrRunningAttributes", ""),
        "2005xx70": ("OW", "1", "SpectrumPixelData", ""),
        "2005xx81": ("UI", "1", "DefaultImageUID", ""),
        "2005xx82": ("CS", "1", "RunningAttributes", ""),
    },
    "Philips MR Imaging DD 004": {
        "2005xx00": ("SS", "1", "MRSpectrumExtraNumber", ""),
        "2005xx01": ("SS", "1", "MRSpectrumKxCoordinate", ""),
        "2005xx02": ("SS", "1", "MRSpectrumKyCoordinate", ""),
        "2005xx03": ("SS", "1", "MRSpectrumLocationNumber", ""),
        "2005xx04": ("SS", "1", "MRSpectrumMixNumber", ""),
        "2005xx05": ("SS", "1", "MRSpectrumXCoordinate", ""),
        "2005xx06": ("SS", "1", "MRSpectrumYCoordinate", ""),
        "2005xx07": ("FL", "1", "MRSpectrumDCLevel", ""),
        "2005xx08": ("FL", "1", "MRSpectrumNoiseLevel", ""),
        "2005xx09": ("FL", "1", "MRSpectrumBeginTime", ""),
        "2005xx10": ("FL", "1", "MRSpectrumEchoTime", ""),
        "2005xx12": ("FL", "1", "SpectrumNumber", ""),
        "2005xx13": ("SS", "1", "MRSpectrumNumber", ""),
        "2005xx14": ("SS", "1", "MRSpectrumNumberOfAverages", ""),
        "2005xx15": ("SS", "1", "MRSpectrumNumberOfSamples", ""),
        "2005xx16": ("SS", "1", "MRSpectrumScanSequenceNumber", ""),
        "2005xx17": ("SS", "1", "MRSpectrumNumberOfPeaks", ""),
        "2005xx18": ("SQ", "1", "MRSpectrumPeak", ""),
        "2005xx19": ("FL", "1", "MRSpectrumPeakIntensity", ""),
        "2005xx20": ("LO", "1", "MRSpectrumPeakLabel", ""),
        "2005xx21": ("FL", "1", "MRSpectrumPeakPhase", ""),
        "2005xx22": ("FL", "1", "MRSpectrumPeakPosition", ""),
        "2005xx23": ("CS", "1", "MRSpectrumPeakType", ""),
        "2005xx24": ("FL", "1", "MRSpectrumPeakWidth", ""),
        "2005xx25": ("CS", "1", "MRSpectroSIB0Correction", ""),
        "2005xx26": ("FL", "1", "MRSpectroB0EchoTopPosition", ""),
        "2005xx27": ("CS", "1", "MRSpectroComplexComponent", ""),
        "2005xx28": ("CS", "1", "MRSpectroDataOrigin", ""),
        "2005xx29": ("FL", "1", "MRSpectroEchoTopPosition", ""),
        "2005xx30": ("CS", "1", "MRInPlaneTransforms", ""),
        "2005xx31": ("SS", "1", "MRNumberOfSpectraAcquired", ""),
        "2005xx33": ("FL", "1", "MRPhaseEncodingEchoTopPositions", ""),
        "2005xx34": ("CS", "1", "MRPhysicalQuantityForChemicalShift", ""),
        "2005xx35": ("CS", "1", "MRPhysicalQuantitySpatial", ""),
        "2005xx36": ("FL", "1", "MRReferenceFrequency", ""),
        "2005xx37": ("FL", "1", "MRSampleOffset", ""),
        "2005xx38": ("FL", "1", "MRSamplePitch", ""),
        "2005xx39": ("SS", "1", "MRSearchIntervalForPeaks", ""),
        "2005xx40": ("CS", "1", "MRSignalDomainForChemicalShift", ""),
        "2005xx41": ("CS", "1", "MRSignalDomainSpatial", ""),
        "2005xx42": ("CS", "1", "MRSignalType", ""),
        "2005xx43": ("CS", "1", "MRSpectroAdditionalRotations", ""),
        "2005xx44": ("SS", "1", "MRSpectroDisplayRanges", ""),
        "2005xx45": ("CS", "1", "MRSpectroEchoAcquisition", ""),
        "2005xx46": ("CS", "1", "MRSpectroFrequencyUnit", ""),
        "2005xx47": ("FL", "1", "MRSpectroGamma", ""),
        "2005xx48": ("CS", "1", "MRSpectroHiddenLineRemoval", ""),
        "2005xx49": ("FL", "1", "MRSpectroHorizontalShift", ""),
        "2005xx50": ("FL", "1", "MRSpectroHorizontalWindow", ""),
        "2005xx51": ("SS", "1", "MRSpectroNumberOfDisplayRanges", ""),
        "2005xx52": ("SS", "1", "MRSpectroNumberOfEchoPulses", ""),
        "2005xx53": ("LO", "1", "MRSpectroProcessingHistory", ""),
        "2005xx54": ("CS", "1", "MRSpectroScanType", ""),
        "2005xx55": ("FL", "1", "MRSpectroSICSIntervals", ""),
        "2005xx56": ("CS", "1", "MRSpectroSIMode", ""),
        "2005xx57": ("SS", "1", "MRSpectroSpectralBW", ""),
        "2005xx58": ("LO", "1", "MRSpectroTitleLine", ""),
        "2005xx59": ("FL", "1", "MRSpectroTurboEchoSpacing", ""),
        "2005xx60": ("FL", "1", "MRSpectroVerticalShift", ""),
        "2005xx61": ("FL", "1", "MRSpectroVerticalWindow", ""),
        "2005xx62": ("FL", "1", "MRSpectroOffset", ""),
        "2005xx63": ("FL", "1", "MRSpectrumPitch", ""),
        "2005xx64": ("CS", "1", "MRVolumeSelection", ""),
        "2005xx70": ("SS", "1", "MRNoMixesSpectro", ""),
        "2005xx71": ("SQ", "1", "MRSeriesSPMix", ""),
        "2005xx72": ("SS", "1", "SPMixTResolution", ""),
        "2005xx73": ("SS", "1", "SPMixKXResolution", ""),
        "2005xx74": ("SS", "1", "SPMixKYResolution", ""),
        "2005xx75": ("SS", "1", "SPMixFResolution", ""),
        "2005xx76": ("SS", "1", "SPMixXResolution", ""),
        "2005xx77": ("SS", "1", "SPMixYResolution", ""),
        "2005xx78": ("SS", "1", "SPMixNoSpectraIntended", ""),
        "2005xx79": ("SS", "1", "SPMixNoAverages", ""),
        "2005xx80": ("SL", "1", "MRSeriesNrOfMFImageObjects", ""),
        "2005xx81": ("IS", "1", "MRScanoGramSurveyNumberOfImages", ""),
        "2005xx82": ("UL", "1", "MRNumberOfProcedureCodes", ""),
        "2005xx83": ("CS", "1", "SortAttributes", ""),
        "2005xx84": ("SS", "1", "MRNrSortAttributes", ""),
        "2005xx85": ("CS", "1", "ImageDisplayDirection", ""),
        "2005xx86": ("CS", "1", "InsetScanogram", ""),
        "2005xx87": ("SS", "1", "MRDisplayLayoutNrColumns", ""),
        "2005xx88": ("SS", "1", "MRDisplayLayoutNrRows", ""),
        "2005xx89": ("SQ", "1", "ViewingProtocol", ""),
        "2005xx90": ("CS", "1", "MRStackCoilFunction", ""),
        "2005xx91": ("PN", "1", "PatientNameJobInParams", ""),
        "2005xx92": ("IS", "1", "MRGeolinkID", ""),
        "2005xx93": ("IS", "1", "MRStationNo", ""),
        "2005xx94": ("CS", "1", "ProcessingHistory", ""),
        "2005xx95": ("UI", "1", "ViewProcedureString", ""),
        "2005xx96": ("CS", "1", "MRFlowImagesPresent", ""),
        "2005xx97": ("LO", "1", "AnatomicRegCodeValue", ""),
        "2005xx98": ("CS", "1", "MRMobiviewEnabled", ""),
        "2005xx99": ("CS", "1", "MRIViewBoldEnabled", ""),
    },
    "Philips MR Imaging DD 005": {
        "2005xx00": ("CS", "1", "MRVolumeViewEnabled", ""),
        "2005xx01": ("UL", "1", "MRNumberOfStudyReference", ""),
        "2005xx03": ("UL", "1", "MRNumberOfSPSCodes", ""),
        "2005xx07": ("SS", "1", "MRNrOfSpecificCharacterSet", ""),
        "2005xx09": ("DS", "1", "RescaleInterceptOriginal", ""),
        "2005xx0a": ("DS", "1", "RescaleSlopeOriginal", ""),
        "2005xx0b": ("LO", "1", "RescaleTypeOriginal", ""),
        "2005xx0e": ("SQ", "1", "PrivateSharedSq", ""),
        "2005xx0f": ("SQ", "1", "PrivatePerFrameSq", ""),
        "2005xx10": ("IS", "1", "MFConvTreatSpectorMixNo", ""),
        "2005xx11": ("UI", "1", "MFPrivateReferencedSOPInstanceUID", ""),
        "2005xx12": ("IS", "1", "MRImageDiffBValueNumber", ""),
        "2005xx13": ("IS", "1", "MRImageGradientOrientationNumber", ""),
        "2005xx14": ("SL", "1", "MRSeriesNrOfDiffBValues", ""),
        "2005xx15": ("SL", "1", "MRSeriesNrOfDiffGradOrients", ""),
        "2005xx16": ("CS", "1", "MRSeriesPlanMode", ""),
        "2005xx17": ("FD", "1", "DiffusionBMatrix", ""),
        "2005xx18": ("CS", "1", "PrivOperatingModeType", ""),
        "2005xx19": ("CS", "1", "PrivOperatingMode", ""),
        "2005xx1a": ("CS", "1", "MRFatSaturationTechnique", ""),
        "2005xx1b": ("IS", "1", "MRVersionNumberDeletedImages", ""),
        "2005xx1c": ("IS", "1", "MRVersionNumberDeletedSpectra", ""),
        "2005xx1d": ("IS", "1", "MRVersionNumberDeletedBlobsets", ""),
        "2005xx1e": ("UL", "1", "LUT1Offset", ""),
        "2005xx1f": ("UL", "1", "LUT1Range", ""),
        "2005xx20": ("UL", "1", "LUT1BeginColor", ""),
        "2005xx21": ("UL", "1", "LUT1EndColor", ""),
        "2005xx22": ("UL", "1", "LUT2Offset", ""),
        "2005xx23": ("UL", "1", "LUT2Range", ""),
        "2005xx24": ("UL", "1", "LUT2BeginColor", ""),
        "2005xx25": ("UL", "1", "LUT2EndColor", ""),
        "2005xx26": ("CS", "1", "ViewingHardcopyOnly", ""),
        "2005xx27": ("SQ", "1", "PrivateEMR", ""),
        "2005xx28": ("SL", "1", "MRSeriesNrOfLabelTypes", ""),
        "2005xx29": ("CS", "1", "MRImageLabelType", ""),
        "2005xx2a": ("CS", "1", "ExamPrintStatus", ""),
        "2005xx2b": ("CS", "1", "ExamExportStatus", ""),
        "2005xx2c": ("CS", "1", "ExamStorageCommitStatus", ""),
        "2005xx2d": ("CS", "1", "ExamMediaWriteStatus", ""),
        "2005xx2e": ("FL", "1", "MRSeriesDBdt", ""),
        "2005xx2f": ("FL", "1", "MRSeriesProtonSAR", ""),
        "2005xx30": ("FL", "1", "MRSeriesNonProtonSAR", ""),
        "2005xx31": ("FL", "1", "MRSeriesLocalSAR", ""),
        "2005xx32": ("CS", "1", "MRSeriesSafetyOverrideMode", ""),
        "2005xx33": ("DT", "1", "EVDVDJobInParamsDatetime", ""),
        "2005xx34": ("DT", "1", "DVDJobInParamsVolumeLabel", ""),
        "2005xx35": ("CS", "1", "SpectroExamcard", ""),
        "2005xx36": ("UI", "1", "MRRefSeriesInstanceUID", ""),
        "2005xx37": ("CS", "1", "ColorLUTType", ""),
        "2005xx3b": ("CS", "1", "MRIsCoilSurvey", ""),
        "2005xx3c": ("FL", "1", "MRStackTablePosLong", ""),
        "2005xx3d": ("FL", "1", "MRStackTablePosLat", ""),
        "2005xx3e": ("FL", "1", "MRStackPosteriorCoilPos", ""),
        "2005xx3f": ("CS", "1", "AIMDLimitsApplied", ""),
        "2005xx40": ("FL", "1", "AIMDHeadSARLimit", ""),
        "2005xx41": ("FL", "1", "AIMDWholeBodySARLimit", ""),
        "2005xx42": ("FL", "1", "AIMDB1RMSLimit", ""),
        "2005xx43": ("FL", "1", "AIMDdbDtLimit", ""),
        "2005xx44": ("IS", "1", "TFEFactor", ""),
        "2005xx45": ("CS", "1", "AttenuationCorrection", ""),
        "2005xx46": ("FL", "1", "FWHMShim", ""),
        "2005xx47": ("FL", "1", "PowerOptimization", ""),
        "2005xx48": ("FL", "1", "CoilQ", ""),
        "2005xx49": ("FL", "1", "ReceiverGain", ""),
        "2005xx4a": ("FL", "1", "DataWindowDuration", ""),
        "2005xx4b": ("FL", "1", "MixingTime", ""),
        "2005xx4c": ("FL", "1", "FirstEchoTime", ""),
        "2005xx4d": ("CS", "1", "IsB0Series", ""),
        "2005xx4e": ("CS", "1", "IsB1Series", ""),
        "2005xx4f": ("CS", "1", "VolumeSelect", ""),
        "2005xx50": ("SS", "1", "MRNrOfPatientOtherIDs", ""),
        "2005xx51": ("IS", "1", "PrivateSeriesNumber", ""),
        "2005xx52": ("UI", "1", "PrivateSeriesInstanceUID", ""),
        "2005xx53": ("CS", "1", "SplitSeriesJobParams", ""),
        "2005xx54": ("SS", "1", "PreferredDimensionForSplitting", ""),
        "2005xx55": ("FD", "1", "ImageVelocityEncodingDirection", ""),
        "2005xx56": ("SS", "1", "ContrastBolusNoInjections", ""),
        "2005xx57": ("LT", "1", "ContrastBolusAgentCode", ""),
        "2005xx58": ("LT", "1", "ContrastBolusAdminRouteCode", ""),
        "2005xx59": ("DS", "1", "ContrastBolusVolume", ""),
        "2005xx5a": ("DS", "1", "ContrastBolusIngredientConcentration", ""),
        "2005xx5b": ("IS", "1", "ContrastBolusDynamicNumber", ""),
        "2005xx5c": ("SQ", "1", "SeriesBolusContrast", ""),
        "2005xx5d": ("IS", "1", "ContrastBolusID", ""),
        "2005xx60": ("CS", "1", "LUTtoRGBJobParams", ""),
        "2005xx92": ("FL", "1", "SpecificEnergyDose", ""),
    },
    "Philips MR Imaging DD 006": {
        "2005xx53": ("FL", "1", "MREFrequency", ""),
        "2005xx54": ("FL", "1", "MREAmplitude", ""),
        "2005xx55": ("FL", "1", "MREMEGFrequency", ""),
        "2005xx56": ("FL", "1", "MREMEGPairs", ""),
        "2005xx57": ("CS", "1", "MREMEGDirection", ""),
        "2005xx58": ("FL", "1", "MREMEGAmplitude", ""),
        "2005xx59": ("FL", "1", "MRENumberofPhaseDelays", ""),
        "2005xx60": ("IS", "1", "MRENumberofMotionCycles", ""),
        "2005xx61": ("FL", "1", "MREMotionMegPhaseDelay", ""),
        "2005xx62": ("LT", "1", "MREInversionAlgorithmVersion", ""),
        "2005xx63": ("CS", "1", "SagittalSliceOrder", ""),
        "2005xx64": ("CS", "1", "CoronalSliceOrder", ""),
        "2005xx65": ("CS", "1", "TransversalSliceOrder", ""),
        "2005xx66": ("CS", "1", "SeriesOrientation", ""),
        "2005xx67": ("IS", "1", "MRStackReverse", ""),
        "2005xx68": ("IS", "1", "MREPhaseDelayNumber", ""),
        "2005xx71": ("IS", "1", "NumberOfInversionDelays", ""),
        "2005xx72": ("FL", "1", "InversionDelayTime", ""),
        "2005xx73": ("IS", "1", "InversionDelayNumber", ""),
        "2005xx74": ("DS", "1", "MaxDbDt", ""),
        "2005xx75": ("DS", "1", "MaxSAR", ""),
        "2005xx76": ("LT", "1", "SARType", ""),
        "2005xx78": ("CS", "1", "MetalImplantStatus", ""),
        "2005xx79": ("CS", "1", "OrientationMirrorFlip", ""),
        "2005xx81": ("CS", "1", "SAROperationMode", ""),
        "2005xx82": ("IS", "1", "SpatialGradient", ""),
        "2005xx83": ("LT", "1", "AdditionalConstraints", ""),
        "2005xx87": ("DS", "1", "PIIM_MR_STUDY_B1RMS", ""),
        "2005xx92": ("SQ", "1", "ContrastInformationSequence", ""),
        "2005xx95": ("CS", "1", "Diffusion2KDTI", ""),
        "2005xx96": ("IS", "1", "DiffusionOrder", ""),
        "2005xx97": ("CS", "1", "IsJEditingSeries ", ""),
        "2005xx98": ("SS", "1 ", "MRSpectrumEditingType", ""),
        "2005xx99": ("SL", "1", "MRSeriesNrOfDiffOrder", ""),
    },
    "SIEMENS MED": {
        "7001xx10": ("LT", "1", "Dummy", ""),
        "7003xx10": ("LT", "1", "Header", ""),
        "7005xx10": ("LT", "1", "Dummy", ""),
    },
    "SIEMENS MED BRAIN ORIENTATION DATA": {
        "0027xx06": ("CS", "1", "Brain Orientation Value", "")
    },
    "SIEMENS MED MEASUREMENT": {
        "0027xx01": ("DS", "3", "Percist Cylinder Position", ""),
        "0027xx02": ("DS", "3", "Percist Cylinder Axis", ""),
        "0027xx03": ("DS", "1", "Percist Cylinder Radius", ""),
        "0027xx04": ("LT", "1", "Isocontour Threshold", ""),
        "0027xx05": ("LO", "1", "Auto Created", ""),
        "0027xx06": ("CS", "1", "Finding Creation Mode", ""),
        "0027xx07": ("DS", "1", "Pet Segmentation Threshold", ""),
        "0027xx08": ("DS", "12", "Change Rate", ""),
        "0027xx09": ("DS", "1", "Volume Doubling Time", ""),
        "0027xx10": ("OB", "1-n", "Change Rates", ""),
    },
    "SIEMENS MED MI": {
        "0067xx01": ("LT", "1", "MI Scan ID", ""),
        "0067xx02": ("LO", "1", "Scanner Console Generation", ""),
        "0067xx03": ("OB", "1-n", "Recon Parameters", ""),
        "0067xx04": ("LO", "1", "Group Reconstruction ID", ""),
        "0067xx05": ("LO", "1", "Device IVK", ""),
        "0067xx14": ("LO", "1", "Raw Data Description", ""),
        "0067xx16": ("UI", "1-n", "Raw Data Series Instance UIDs", ""),
        "0067xx17": ("UI", "1-n", "Raw Data Referenced Series Instance UIDs", ""),
    },
    "SIEMENS MED ORIENTATION RESULT": {
        "0027xx05": ("CS", "1", "Cardiac Orientation Value", "")
    },
    "SIEMENS MED PT": {
        "0071xx21": ("UI", "1", "Reference To Registration", ""),
        "0071xx22": ("DT", "1", "Decay Correction DateTime", ""),
        "0071xx23": ("US", "16", "Registration Matrix", ""),
        "0071xx24": ("CS", "1", "Table Motion", ""),
        "0071xx25": ("FD", "1", "Lumped Constant", ""),
        "0071xx26": ("CS", "1", "Histogramming Method", ""),
    },
    "SIEMENS MED PT MU MAP": {
        "0071xx01": ("UI", "1", "SOP Class of Source", ""),
        "0071xx02": ("UI", "1", "Related Mu Map Series", ""),
    },
    "SIEMENS MED RTSTRUCT": {"0063xx32": ("DS", "3", "GTV Marker Position", "")},
    "SIEMENS MI RWVM SUV": {"0041xx01": ("CS", "1", "SUV Decay Correction Method", "")},
    "SIEMENS PET DERIVED": {
        "0075xx01": ("US", "1", "Volume Index", ""),
        "0075xx02": ("IS", "1", "Time Slice Duration", ""),
        "0075xx03": ("SQ", "1", "Frame Reference Time Sequence", ""),
    },
    "SIEMENS RA GEN": {
        "0011xx26": ("SL", "1", "PtopTotalSkinDose", ""),
        "0019xx1f": ("SS", "1", "DefaultTableIsoCenterHeight", ""),
        "0019xxa2": ("SL", "1", "SceneNumber", ""),
        "0019xxa4": ("SS", "1", "AcquisitionMode", ""),
        "0019xxa5": ("SS", "1", "AcquisitonFrameRate", ""),
        "0019xxa6": ("SL", "1", "ECGFlag", ""),
        "0019xxa7": ("SL", "1", "AdditionalSceneData", ""),
        "0019xxa8": ("SL", "1", "FileCopyFlag", ""),
        "0019xxa9": ("SL", "1", "PhlebovisionFlag", ""),
        "0019xxaa": ("SL", "1", "Co2Flag", ""),
        "0019xxab": ("SS", "1", "MaxSpeed", ""),
        "0019xxac": ("SS", "1", "StepWidth", ""),
        "0019xxad": ("SL", "1", "DigitalAcquisitionZoom", ""),
        "0019xxff": ("SS", "1-n", "Internal", ""),
        "0021xx27": ("SS", "1", "PlaneBImagesExist", ""),
        "0021xx28": ("SS", "1", "NoOf2MBChunks", ""),
        "0021xx40": ("SS", "1", "ArchiveSWInternalVersion", ""),
    },
    "SIEMENS RA PLANE A": {
        "0011xx28": ("SL", "1", "FluoroTimerA", ""),
        "0011xx29": ("SL", "1", "FluoroSkinDoseA", ""),
        "0011xx2a": ("SL", "1", "TotalSkinDoseA", ""),
        "0011xx2b": ("SL", "1", "FluoroDoseAreaProductA", ""),
        "0011xx2c": ("SL", "1", "TotalDoseAreaProductA", ""),
        "0019xx15": ("LT", "1", "OfflineUID", ""),
        "0019xx18": ("SS", "1", "Internal", ""),
        "0019xx19": ("SS", "1", "Internal", ""),
        "0019xx1a": ("SS", "1", "Internal", ""),
        "0019xx1b": ("SS", "1", "Internal", ""),
        "0019xx1c": ("SS", "1", "Internal", ""),
        "0019xx1d": ("SS", "1", "Internal", ""),
        "0019xx1e": ("SS", "1", "Internal", ""),
        "0019xx1f": ("SS", "1-n", "Internal", ""),
        "0019xx2a": ("SS", "1", "AcquisitionDelay", ""),
        "0019xxae": ("SL", "1", "IIToCoverDistance", ""),
        "0019xxb0": ("SS", "1", "LastFramePhase1", ""),
        "0019xxb1": ("SS", "1", "FrameRatePhase1", ""),
        "0019xxb2": ("SS", "1", "LastFramePhase2", ""),
        "0019xxb3": ("SS", "1", "FrameRatePhase2", ""),
        "0019xxb4": ("SS", "1", "LastFramePhase3", ""),
        "0019xxb5": ("SS", "1", "FrameRatePhase3", ""),
        "0019xxb6": ("SS", "1", "LastFramePhase4", ""),
        "0019xxb7": ("SS", "1", "FrameRatePhase4", ""),
        "0019xxb8": ("SS", "1", "GammaOfNativeImage", ""),
        "0019xxb9": ("SS", "1", "GammaOfTVSystem", ""),
        "0019xxbb": ("SL", "1", "PixelshiftX", ""),
        "0019xxbc": ("SL", "1", "PixelshiftY", ""),
        "0019xxbd": ("SL", "1", "MaskAverageFactor", ""),
        "0019xxbe": ("SL", "1", "BlankingCircleFlag", ""),
        "0019xxbf": ("SL", "1", "CircleRowStart", ""),
        "0019xxc0": ("SL", "1", "CircleRowEnd", ""),
        "0019xxc1": ("SL", "1", "CircleColumnStart", ""),
        "0019xxc2": ("SL", "1", "CircleColumnEnd", ""),
        "0019xxc3": ("SL", "1", "CircleDiameter", ""),
        "0019xxc4": ("SL", "1", "RectangularCollimaterFlag", ""),
        "0019xxc5": ("SL", "1", "RectangleRowStart", ""),
        "0019xxc6": ("SL", "1", "RectangleRowEnd", ""),
        "0019xxc7": ("SL", "1", "RectangleColumnStart", ""),
        "0019xxc8": ("SL", "1", "RectangleColumnEnd", ""),
        "0019xxc9": ("SL", "1", "RectangleAngulation", ""),
        "0019xxca": ("SL", "1", "IrisCollimatorFlag", ""),
        "0019xxcb": ("SL", "1", "IrisRowStart", ""),
        "0019xxcc": ("SL", "1", "IrisRowEnd", ""),
        "0019xxcd": ("SL", "1", "IrisColumnStart", ""),
        "0019xxce": ("SL", "1", "IrisColumnEnd", ""),
        "0019xxcf": ("SL", "1", "IrisAngulation", ""),
        "0019xxd1": ("SS", "1", "NumberOfFramesPlane", ""),
        "0019xxd2": ("SS", "1", "Internal", ""),
        "0019xxd3": ("SS", "1", "Internal", ""),
        "0019xxd4": ("SS", "1", "Internal", ""),
        "0019xxd5": ("SS", "1", "Internal", ""),
        "0019xxd6": ("SS", "1-n", "Internal", ""),
        "0019xxd7": ("SS", "1-n", "Internal", ""),
        "0019xxd8": ("SS", "1", "Internal", ""),
        "0019xxd9": ("SS", "1", "Internal", ""),
        "0019xxda": ("SS", "1", "Internal", ""),
        "0019xxdb": ("SS", "1", "Internal", ""),
        "0019xxdc": ("SS", "1", "Internal", ""),
        "0019xxdd": ("SL", "1", "AnatomicBackground", ""),
        "0019xxde": ("SL", "1-n", "AutoWindowBase", ""),
        "0019xxdf": ("SS", "1", "Internal", ""),
        "0019xxe0": ("SL", "1", "Internal", ""),
    },
    "SIEMENS RA PLANE B": {
        "0011xx28": ("SL", "1", "FluoroTimerB", ""),
        "0011xx29": ("SL", "1", "FluoroSkinDoseB", ""),
        "0011xx2a": ("SL", "1", "TotalSkinDoseB", ""),
        "0011xx2b": ("SL", "1", "FluoroDoseAreaProductB", ""),
        "0011xx2c": ("SL", "1", "TotalDoseAreaProductB", ""),
    },
    "SIENET": {"0099xx02": ("UL", "1", "DataObjectAttributes", "")},
    "SONOWAND AS": {
        "0135xx10": ("LO", "1", "UltrasoundScannerName", ""),
        "0135xx11": ("LO", "1", "TransducerSerial", ""),
        "0135xx12": ("LO", "1", "ProbeApplication", ""),
    },
    "SPI-P-Private-DCI Release 1": {
        "0019xx10": ("UN", "1", "ECGTimeMapDataBitsAllocated", ""),
        "0019xx11": ("UN", "1", "ECGTimeMapDataBitsStored", ""),
        "0019xx12": ("UN", "1", "ECGTimeMapDataHighBit", ""),
        "0019xx13": ("UN", "1", "ECGTimeMapDataRepresentation", ""),
        "0019xx14": ("UN", "1", "ECGTimeMapDataSmallestDataValue", ""),
        "0019xx15": ("UN", "1", "ECGTimeMapDataLargestDataValue", ""),
        "0019xx16": ("UN", "1", "ECGTimeMapDataNumberOfDataValues", ""),
        "0019xx17": ("UN", "1", "ECGTimeMapData", ""),
    },
    "SPI-P-Private_ICS Release 1": {
        "0029xx43": ("SQ", "1", "Unknown", ""),
        "0029xx44": ("SQ", "1", "Unknown", ""),
    },
    "SVISION": {"0025xx21": ("UI", "1", "Unknown", "")},
}
